#include "map_test.h"

namespace folly_test {

class FollyMapInsDelFindTest_Parallel : public cds_test::stress_fixture {
protected:
  static unsigned s_nInsertPercentage;
  static unsigned s_nDeletePercentage;
  static size_t s_nThreadCount;
  static size_t s_nMapKeyRange;
  static size_t s_nInitialMapSize;

  enum actions { do_find, do_insert, do_delete };
  static const unsigned int kShuffleSize = 100;
  static actions s_arrShuffle[kShuffleSize];

  static size_t s_nConcurrentHashMapPassCount;
  static size_t s_nAtomicHashMapPassCount;
  static size_t s_nAtomicUnorderedInsertMapPassCount;

  static void InitShuffleArray() {
    // Build an array of shuffled actions.
    EXPECT_LE(s_nInsertPercentage + s_nDeletePercentage, 100);
    actions* pFirst = s_arrShuffle;
    actions* pLast = s_arrShuffle + s_nInsertPercentage;
    std::fill(pFirst, pLast, do_insert);
    pFirst = pLast;
    pLast += s_nDeletePercentage;
    std::fill(pFirst, pLast, do_delete);
    pFirst = pLast;
    pLast = s_arrShuffle + sizeof(s_arrShuffle) / sizeof(s_arrShuffle[0]);
    if (pFirst < pLast) {
      std::fill(pFirst, pLast, do_find);
    }
    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(s_arrShuffle, pLast, g);
  }

  static void SetUpTestCase() {
    InitShuffleArray();
    const cds_test::config& cfg = get_config("ParallelFollyMap");
    GetConfigNonZeroExpected(InsertPercentage, 5);
    GetConfigNonZeroExpected(DeletePercentage, 5);
    GetConfigNonZeroExpected(ThreadCount, 4);
    GetConfigNonZeroExpected(MapKeyRange, 20000);
    GetConfigNonZeroExpected(InitialMapSize, 20000);
    GetConfigNonZeroExpected(ConcurrentHashMapPassCount, 1000);
    GetConfigNonZeroExpected(AtomicHashMapPassCount, 1000);
    GetConfigNonZeroExpected(AtomicUnorderedInsertMapPassCount, 1000);
  }

  template <typename Map>
  static void run_test(Map* map, size_t pass_count, size_t* inserted_num,
                       size_t* deleted_num) {
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<size_t> dis(2, s_nMapKeyRange);

    unsigned action_index = 0;
    size_t nInsertedNum = 0;
    size_t nDeletedNum = 0;
    size_t nFindSuccess = 0;
    size_t nOperations = 0;

    for (size_t count = 0; count < pass_count; count++) {
      // The number to operate on the map.
      size_t key = dis(gen);
      switch (s_arrShuffle[action_index]) {
        case do_insert: {
          size_t val = dis(gen);
          if (map_insert(map, key, val)) {
            nInsertedNum++;
          }
          break;
        }
        case do_delete: {
          if (map_delete(map, key)) {
            nDeletedNum++;
          }
          break;
        }
        case do_find: {
          if (map_find(map, key)) {
            ++nFindSuccess;
          }
          break;
        }
        default: { break; }
      }
      if (++action_index >= kShuffleSize) {
        action_index = 0;
      }
    }
    *inserted_num = nInsertedNum;
    *deleted_num = nDeletedNum;
  }
};

size_t FollyMapInsDelFindTest_Parallel::s_nThreadCount;
size_t FollyMapInsDelFindTest_Parallel::s_nMapKeyRange;
size_t FollyMapInsDelFindTest_Parallel::s_nInitialMapSize;
unsigned FollyMapInsDelFindTest_Parallel::s_nInsertPercentage;
unsigned FollyMapInsDelFindTest_Parallel::s_nDeletePercentage;
const unsigned int FollyMapInsDelFindTest_Parallel::kShuffleSize;
FollyMapInsDelFindTest_Parallel::actions FollyMapInsDelFindTest_Parallel::
    s_arrShuffle[FollyMapInsDelFindTest_Parallel::kShuffleSize];
size_t FollyMapInsDelFindTest_Parallel::s_nConcurrentHashMapPassCount;
size_t FollyMapInsDelFindTest_Parallel::s_nAtomicHashMapPassCount;
size_t FollyMapInsDelFindTest_Parallel::s_nAtomicUnorderedInsertMapPassCount;

#define FollyMapThreading(map_type, pass_count)                                \
  std::unique_ptr<std::thread[]> threads(new std::thread[s_nThreadCount]);     \
  std::unique_ptr<size_t[]> inserted_nums(new size_t[s_nThreadCount]);         \
  std::unique_ptr<size_t[]> deleted_nums(new size_t[s_nThreadCount]);          \
  for (size_t i = 0; i < s_nThreadCount; i++) {                                \
    threads[i] = std::thread(run_test<map_type>, map.get(), pass_count,        \
                             &inserted_nums[i], &deleted_nums[i]);             \
  }                                                                            \
  size_t inserted_sum = 0;                                                     \
  size_t deleted_sum = 0;                                                      \
  for (size_t i = 0; i < s_nThreadCount; i++) {                                \
    threads[i].join();                                                         \
    inserted_sum += inserted_nums[i];                                          \
    deleted_sum += deleted_nums[i];                                            \
  }                                                                            \
  EXPECT_LE(deleted_sum, inserted_sum);

TEST_F(FollyMapInsDelFindTest_Parallel, FollyConcurrentHashMap) {
  std::unique_ptr<ConcurrentHashMap> map(
      new ConcurrentHashMap(s_nInitialMapSize));
  FollyMapThreading(ConcurrentHashMap, s_nConcurrentHashMapPassCount);
}

TEST_F(FollyMapInsDelFindTest_Parallel, FollyAtomicHashMap) {
  std::unique_ptr<AtomicHashMap> map(
      new AtomicHashMap(s_nInitialMapSize));
  FollyMapThreading(AtomicHashMap, s_nAtomicHashMapPassCount);
}

TEST_F(FollyMapInsDelFindTest_Parallel, FollyAtomicUnorderedInsertMap) {
  std::unique_ptr<AtomicUnorderedInsertMap> map(
      new AtomicUnorderedInsertMap(s_nInitialMapSize));
  FollyMapThreading(AtomicUnorderedInsertMap,
     s_nAtomicUnorderedInsertMapPassCount);
}

} // namespace folly_test
